begin
  require 'spreadsheet'
rescue LoadError
  ActionController::Base::logger.info 'Duplicate views controller: spreadsheet gem not available'
end

class ListDuplicateViewsController < ApplicationController
  unloadable
  layout 'base'
  before_filter :require_admin

  def index
    redmine_app_folder=Rails.root
    @list_plugin_folders=true if params[:show_plugins_views] && params[:show_plugins_views] == '1'
    # array of [view name, view files array in this view [filename,array of plugins names]]
    @duplicate_views=get_duplicate_views_list(redmine_app_folder)
    if request.post?
      if params[:save_legend] && defined?(Spreadsheet::Workbook.new)
        send_data(save_legend_to_xls, :type => :xls, :filename => 'redmine_duplicate_views.xls')
      end
    end

  end

  MODELS_WITH_ATTACHMENTS = %w(project version wiki document message news issue)
  MAX_OBJECTS_TO_FIND = 300

  def batch_remove_projects
    if request.post?
      prj_names_list=[]
      prj_error_list=[]
      projects=[]
      params[:remove_ids].each do |p_str|
        projects << p_str.to_i
        project = Project.find_by_id(p_str.to_i)
        projects += project.descendants.collect{|p| p.id} if project.descendants.any?
      end
      projects.uniq!
      projects.each do |pid|
        project = Project.find_by_id(pid)
        if project
          if params[:wipe_attachments]
            MODELS_WITH_ATTACHMENTS.each do |model|
              send("wipe_project_#{model}_files",project,params[:wipe_folder] && params[:wipe_folder] != '' ? params[:wipe_folder] : nil) if respond_to?("wipe_project_#{model}_files",true)
            end
          end
          query = Query.new(:name => "_ldv_query", :project => project)
          query.add_filter("status_id", "*", [""])
          cond=(query.statement + ' AND (issues.parent_id IS NULL)')
          issues = []
          begin
            issues=Issue.find(:all,:conditions => cond, :limit => 300, :include => [ :assigned_to, :project, :fixed_version, :tracker, :status ])
            issues.each do |i|
              i.destroy
            end
          end while issues.size > 0
          prj_names_list << project.name
          issues=Issue.find(:all,:conditions => query.statement, :limit => 300, :include => [ :assigned_to, :project, :fixed_version, :tracker, :status ])
          if issues.size > 0
            logger.info("Not all issues were destroyed!")
            prj_error_list << project.name
          end
        end
      end
      projects.each do |pid|
        project = Project.find_by_id(pid)
        project.destroy if project
      end
      flash_str="Projects deleted:\n"+prj_names_list.join(", ")
      if prj_error_list.size > 0
        flash_str+="\nProjects with issues consistency errors:\n"+prj_error_list.join(", ")
      end
      if prj_error_list.size > 0
        flash[:error]=flash_str
      else
        flash[:notice]=flash_str
      end
      redirect_to :action => 'batch_remove_projects'
      return
    end
    @projects=Project.find(:all,:order => 'lft')
  end

  def clean_up_attachments_folder
    if request.post?
      offset=0
      base_path = Pathname.new(Attachment.storage_path)
      files=Dir[File.join(base_path,"**","*")].collect do |file|
        unless File.directory?(file)
          Pathname.new(file).relative_path_from(base_path)
        else
          ''
        end
      end
      files=files.select do |file|
        true unless file == ''
      end

      attaches=[]
      begin
        attaches=Attachment.find(:all,:limit => MAX_OBJECTS_TO_FIND, :offset => offset)
        attaches.each do |attach|
          idx=files.index(Pathname.new(attach.diskfile).relative_path_from(base_path))
          files[idx]='' if idx
        end
        offset+=MAX_OBJECTS_TO_FIND
      end while attaches.size > 0
      files=files.select do |file|
        true unless file == ''
      end

      logger.info("Found #{files.size} orphaned files\n")
      files.each do |file|
        full_fname=File.join(base_path,file)
        params[:wipe_folder] && params[:wipe_folder] != '' ?  move_attachment(full_fname,params[:wipe_folder]) : File.delete(full_fname)
      end

      flash[:notice]="#{files.size} orphaned files were moved."

      redirect_to :action => 'clean_up_attachments_folder'
      return
    end
  end

private

  REDMINE_MAIN_APP_NAME = '_redmine_main_app'

  def get_duplicate_views_list(root_path)
    views_files={}

    views_folder="#{root_path}/app/views/"
    views_list=Dir.glob("#{views_folder}*").collect do |d|
      File.basename(d)
    end
    views_list.each do |v|
      views_files[v]={}
      process_view_folder(views_folder,views_folder,v,views_files[v])
    end

    plugins_folder=(Redmine::VERSION::MAJOR >= 2 ? "#{root_path}/plugins/" : "#{root_path}/vendor/plugins/")
    plugins_list=Dir.glob("#{plugins_folder}*").collect do |d|
      File.basename(d)
    end

    plugins_list.each do |p|
      process_plugin_folder(views_files,p,plugins_folder)
    end

    dup_views=views_files.select do |view,dups|
      r=false
      dups.each_pair do |vf,pa|
        r=true if pa.size > 1
      end
      r
    end

    dup_views.reject! do |e|
      e[1]=e[1].to_a

      e[1].reject! do |a|
        r=false
        if a[1].size > 1
          a[1].reject! do |ra|
            true if ra == REDMINE_MAIN_APP_NAME
          end
        else
          r=true
        end
        r
      end

      e[1].sort! do |a,b|
        if a[1].size > 1 && b[1].size < 2
          -1
        else
          if b[1].size > 1 && a[1].size < 2
           1
          else
           if b[1].size == a[1].size
             a[0] <=> b[0]
           else
             a[1].size <=> b[1].size
           end
          end
        end
      end
      r=false
      r=true unless e[1].size > 0
      r
    end

    dup_views.sort! do |a,b|
        ea=a[1].detect do |e|
          e[1].size > 1
        end
        eb=b[1].detect do |e|
          e[1].size > 1
        end
        if (ea == nil && eb == nil) || (ea != nil && eb != nil && ea[1].size == eb[1].size)
          a[0] <=> b[0]
        else
          if ea == nil
            1
          else
            if eb == nil
              -1
            else
              ea[1].size == eb[1].size
            end
          end
        end
    end

    return dup_views
  end

  def process_view_folder(root_dir,base_dir,view_dir,view_files)
    Dir.glob("#{base_dir}#{view_dir}/*").each do |d|
      if File.directory?(d)
        process_view_folder(root_dir,"#{base_dir}#{view_dir}/",File.basename(d),view_files)
      else
        view_files[strip_extra_extensions(d.gsub(root_dir,''))]=[REDMINE_MAIN_APP_NAME]
      end
    end
  end

  def process_plugin_folder(views_files,plugin_name,plugins_folder)
    pv_folder="#{plugins_folder}#{plugin_name}/app/views/"
    pv_files={}
    pv_list=Dir.glob("#{pv_folder}*").collect do |d|
      File.basename(d)
    end

    pv_list.each do |v|
      pv_files[v]={}
      process_view_folder(pv_folder,pv_folder,v,pv_files[v])
    end

    pv_files.each_pair do |view_name,data|
      if views_files[view_name]
        data.each_pair do |view_file,data2|
          view_stripped=strip_extra_extensions(view_file)
          if views_files[view_name][view_stripped] || @list_plugin_folders
            views_files[view_name][view_stripped] = [] unless views_files[view_name][view_stripped]
            views_files[view_name][view_stripped] << plugin_name
          end
        end
      else
        if @list_plugin_folders
          views_files[view_name]={}
          data.each_pair do |view_file,data2|
            view_stripped=strip_extra_extensions(view_file)
            views_files[view_name][view_stripped] = [] unless views_files[view_name][view_stripped]
            views_files[view_name][view_stripped] << plugin_name
          end
        end
      end
    end
  end

  def save_legend_to_xls
    book = Spreadsheet::Workbook.new

    xcf={}
    xcf[:header_1]=Spreadsheet::Format.new :weight => :bold, :size => 24
    xcf[:header_2]=Spreadsheet::Format.new :size => 20
    xcf[:table_elem]=Spreadsheet::Format.new
    xcf[:table_elem].border=true
    xcf[:table_elem_bold]=Spreadsheet::Format.new :weight => :bold
    xcf[:table_elem_bold].border=true
    xcf[:table_elem_red]=Spreadsheet::Format.new :color => :red
    xcf[:table_elem_red].border=true

    sheet1 = book.create_worksheet(:name => 'Views list (table)')

    sheet1.row(0).replace [l(:table_ldv_header_view_name),l(:table_ldv_header_dup_filename),l(:table_ldv_header_plugin_name)]
    3.times do |i|
      sheet1.row(0).set_format(i,xcf[:table_elem_bold])
    end
    current_row=1

    @duplicate_views.each do |dv|
      dv[1].each do |v|
        v[1].each do |n|
          sheet1.row(current_row)[0]=dv[0]
          sheet1.row(current_row)[1]=v[0]
          sheet1.row(current_row)[2]=n
          3.times do |i|
            sheet1.row(current_row).set_format(i,xcf[v[1].size > 1 ? :table_elem_red : :table_elem])
          end
          current_row+=1
        end
      end
    end

    #sheet1.row(current_row)[0]="Text1\nText2\nText3"

    [30,50,50].each_with_index do |w,idx|
      sheet1.column(idx).width=w
      #sheet1.column(idx).default_format=xcf[:word_wrap_format]
    end

    sheet1 = book.create_worksheet(:name => 'Views list')

    current_row=0

    @duplicate_views.each do |dv|
      sheet1.row(current_row)[0]=dv[0]
      sheet1.row(current_row).set_format(0,xcf[:header_1])
      current_row+=1
      dv[1].each do |v|
        sheet1.row(current_row)[0]=v[0]
        sheet1.row(current_row).set_format(0,xcf[:header_2])
        current_row+=1
        v[1].each do |n|
          sheet1.row(current_row)[0]=n
          current_row+=1
        end
      end
      current_row+=1
    end


    xls_stream = StringIO.new('')
    book.write(xls_stream)

    return xls_stream.string
  end

  def strip_extra_extensions(view_name)
    md=view_name.match(/(.*)(\.rhtml|\.html\.erb)\z/)
    return md[1] if md
    view_name
  end

  def move_attachment(attach_filename,path_to_move)
    fnm=File.basename(attach_filename)
    FileUtils.makedirs("#{path_to_move}/#{fnm[0..3]}")
    if File.exist?(attach_filename)
      begin
        FileUtils.mv(attach_filename,"#{path_to_move}/#{fnm[0..3]}/#{fnm}")
      rescue
        logger.info("Exception while moving file #{attach_filename}\n")
      end
    end
  end

  def wipe_attachments_from_disk(attachments,path_to_move)
    attachments.each do |attach|
      logger.info("Removing attachment ID: #{attach.id}, Filename: #{attach.diskfile}\n")
      path_to_move ? move_attachment(attach.diskfile,path_to_move) : File.delete(attach.diskfile)
    end
  end

  def wipe_project_project_files(project,path_to_move)
    logger.info("Wiping files attached to project #{project.name}. Total: #{project.attachments.size}\n")
    wipe_attachments_from_disk(project.attachments,path_to_move)
  end

  def wipe_project_version_files(project,path_to_move)
    versions=Version.find(:all,:conditions => ("versions.project_id IN (#{project.id})"), :include => [ :attachments ])
    logger.info("Wiping files attached to versions of project #{project.name}. Found #{versions.size} versions\n")
    versions.each do |v|
      wipe_attachments_from_disk(v.attachments,path_to_move)
    end
  end

  def wipe_project_document_files(project,path_to_move)
    objs=Document.find(:all,:conditions => ("documents.project_id IN (#{project.id})"), :include => [ :attachments ])
    logger.info("Wiping files attached to documents of project #{project.name}. Found #{objs.size} documents\n")
    objs.each do |v|
      wipe_attachments_from_disk(v.attachments,path_to_move)
    end
  end

  def wipe_project_wiki_files(project,path_to_move)
    objs=Wiki.find(:all,:conditions => ("wikis.project_id IN (#{project.id})"), :include => [ :pages ])
    logger.info("Wiping files attached to wikis of project #{project.name}. Found #{objs.size} wikis\n")
    objs.each do |v|
      v.pages.each do |page|
        logger.info("Wiping files attached to wiki page #{page.title} of project #{project.name}. Total #{page.attachments.size} files\n")
        wipe_attachments_from_disk(page.attachments,path_to_move)
      end
    end
  end

  def wipe_project_message_files(project,path_to_move)
    objs=Board.find(:all,:conditions => ("boards.project_id IN (#{project.id})"), :include => [ :messages, :topics ])
    logger.info("Wiping files attached to boards of project #{project.name}. Found #{objs.size} boards\n")
    objs.each do |v|
      logger.info("Wiping files attached to messages of board #{v.name} of project #{project.name}.\n")
      v.messages.each do |msg|
        wipe_attachments_from_disk(msg.attachments,path_to_move)
      end
    end
  end

  def wipe_project_news_files(project,path_to_move)
    objs=News.find(:all,:conditions => ("news.project_id IN (#{project.id})"), :include => [ :attachments ])
    logger.info("Wiping files attached to news of project #{project.name}. Found #{objs.size} news\n")
    objs.each do |v|
      wipe_attachments_from_disk(v.attachments,path_to_move)
    end
  end

  def wipe_project_issue_files(project,path_to_move)
    logger.info("Wiping files attached to issues of project #{project.name}.\n")
    objs=[]
    offset=0
    begin
      objs=Issue.find(:all,:limit => MAX_OBJECTS_TO_FIND, :offset => offset,:conditions => ("issues.project_id IN (#{project.id})"), :include => [ :attachments ])
      objs.each do |v|
        wipe_attachments_from_disk(v.attachments,path_to_move)
      end
      offset+=MAX_OBJECTS_TO_FIND
    end while objs.size > 0
  end

end

